import logging
import os
import re
from itertools import compress
from time import sleep, time
from datetime import datetime

from serial import SerialException

from ziva.jda_to_dtv import Jda2dtv
from .comm_package import validate_data, define_package, data_to_parsed_string, replace_repeating_chars, data_decompress
from .serialib import SerialComm
from .const import *
from .exceptions import (
    NoAnswer, InvalidAnswer,
    InvalidCommand,
    CRCError,
    DeviceNotInitialized,
    ParameterError,
    DeviceTimeout,
    InvalidData, InvalidAddress
)

logger = logging.getLogger(__name__)


class Ziva(SerialComm):
    def __init__(self):
        super().__init__()
        # Serial port
        self.port = None
        self.baudrate = 38400
        self.timeout = 1
        self.package_counter = 64

        self.recv_address = None
        self.rf = None
        self.wake_up_time = WAKE_UP_TIME_TOTAL

        self.retry_limit = 5

        # Real values (actual values from device)
        self.rv_units = []
        self.rv_names = []
        self.real_values = []

        # info values, user values and memory status
        # Latest data combines all of the above with last
        self.info_values = {}
        self.user_values = {}
        self.memory_status = None
        self.data = {
            'real_values': self.real_values,
            'info_params': self.info_values,
            'user_params': self.user_values,
            'memory_status': self.memory_status,
        }

        self.parameters = {}  # dict of all parameters that are read
        self.settings = []

        self.memory_data = None

        # Application commands for communication
        self.real_values_params = [RV_MASK, RV_UNITS, RV_NAMES]
        self.info_params = [IDENT_A, IDENT_B, IDENT_C, ID_1]
        self.user_params = []

        # Information for reading memory data
        self.total_data_len = 0
        self.total_data_read = 0
        self.read_memory_progress = 0

        self.raw_memory_data = []

    def __update_param(self, name, value):
        self.parameters[name] = value
        # also update in settings
        item = [i for i in self.settings if i['name'] == name]
        if item:
            item[0]['value'] = value

    def __increment_counter(self):
        """Counter. Increment counter after every sent packet"""

        if self.package_counter == 127:
            self.package_counter = 64
        else:
            self.package_counter += 1

    def __define_package(self, app_cmd: str = None, var_name: str = None, var_val: str = None):
        """Define package
        :param app_cmd: Application command (VR (Value read), VW (Value write),...)
        :param var_name: Variable name  (RV, IDENT_A, IDENT_B,..)
        :param var_val: Variable value (string that can represent anything)
        """

        package = define_package(
            recv_addr=self.recv_address,
            app_cmd=app_cmd,
            var_name=var_name,
            var_val=var_val,
            counter=self.package_counter,
            rf=self.rf
        )
        return package

    def initialize(self, recv_address: int, rf: bool = True, wake_up_time: int = WAKE_UP_TIME_TOTAL,
                   user_params: list = None):
        """
        Set device. Call this on start of the communication.

        ISM (non RF) devices can have recv_addr from 1-255; 255 is used as a universal address (wakes up all devices
        regardless of the actual device address)

        :param recv_address: receiver address (eg 1 for ISM or 1019 for RF)
        :param rf: RF or ISM
        :param wake_up_time: Wake up time for RF devices
        :param user_params : Additional user params that are called when connecting to device
        """

        self.recv_address = recv_address
        self.rf = rf
        self.user_params = []
        self.user_values = {}
        self.wake_up_time = wake_up_time if wake_up_time else WAKE_UP_TIME_TOTAL
        if self.rf:
            self.user_params.append(RECORD_TIME)
            self.timeout = 1
        else:
            if recv_address < 1 or recv_address > 255:
                raise InvalidAddress('Address of the device is not valid')
            self.timeout = 0.5
            self.retry_limit = 3
        if user_params:
            self.user_params += user_params

    def deinitialize(self):
        """Deinitialize device."""

        self.__init__()

    def read_variable(self, name: str, timeout: float = None, retry_limit: int = None) -> dict:
        """Read variable

        :param timeout: Timeout if there is no answer from device; if None, default timeout is used
        :param retry_limit: Retry if exception in send_receive routine is raised
        :param name: Name of parameter to read e.g. RECORD_TIME
        """

        if name is None:
            raise ValueError('Name of the variable is not defined')

        data = self.send_receive(app_cmd=VALUE_READ, var_name=name, timeout=timeout, retry_limit=retry_limit)
        self.parameters[name] = data['data']
        return data

    def write_variable(self, name=None, value=None, timeout: float = None, retry_limit: int = None) -> dict:
        """Write variable

        :param timeout: Timeout if there is no answer from device; if None, default timeout is used
        :param retry_limit: Retry if exception in send_receive routine is raised
        :param name: Name of parameter to write e.g. RECORD_TIME
        :param value: Value to write e.g. 60
        :return ??
        """

        if name is None or value is None:
            raise ValueError('Name or value of the variable is not defined')
        name, value = str(name), str(value)
        data = self.send_receive(app_cmd=VALUE_WRITE, var_name=name, var_val=value, timeout=timeout,
                                 retry_limit=retry_limit)
        self.__update_param(name=name, value=value)
        return data

    def send_receive(self, app_cmd: str, var_name: str = None, var_val: str = None, retry_limit: int = None,
                     timeout: float = None, stream_channel: bool = False, parse_to_str: bool = True) -> dict:
        """Sent package and receive answer."""

        if not self.recv_address:
            raise DeviceNotInitialized('Device is not initialized')

        if timeout is None:
            timeout = self.timeout
        if retry_limit is None:
            retry_limit = self.retry_limit

        if self.ser is not None:
            # Update timeout for serial if it changes
            if timeout != self.ser.timeout:
                self.ser.timeout = timeout

        for error_count in range(retry_limit):
            package = self.__define_package(
                app_cmd=app_cmd,
                var_name=var_name,
                var_val=var_val,
            )
            try:
                self.write(data=package)
                data = self.read(last_char=END_MARKER_BYTES)
            except SerialException as e:
                logger.warning(e)
                raise
            try:
                replaced_data = replace_repeating_chars(data, reverse=True)
                validated_data = validate_data(replaced_data, stream_channel=stream_channel)
                if parse_to_str:
                    validated_data['data'] = data_to_parsed_string(data=validated_data['data'])
                self.__increment_counter()
                if stream_channel:
                    # Remember all raw data for testing in case of invalid data decompression
                    self.raw_memory_data.append(data)
            except CRCError as e:
                logger.warning(e)
                self.__increment_counter()
                if retry_limit == error_count + 1:
                    logger.error(e)
                    raise
            except InvalidAnswer as e:
                logger.warning(e)
                self.__increment_counter()
                if stream_channel:
                    # After InvalidAnswer is raised we cannot stream data anymore because data packages are invalid
                    raise InvalidData('Invalid data from device')
                if retry_limit == error_count + 1:
                    logger.error(e)
                    raise
            except (NoAnswer, DeviceTimeout) as e:
                logger.warning(e)
                if retry_limit == error_count + 1:
                    logger.error(e)
                    self.__increment_counter()
                    raise
            except (InvalidCommand, Exception) as e:
                self.__increment_counter()
                logger.warning(e)
                raise
            else:
                return validated_data

    def get_data(self) -> dict:
        """ Get latest data from device and update parameters if they were changed

        :return: dictionary with latest data
        """

        if not self.rf:
            # Adjust retry limit for retrieving info params
            # Only to lower the wait time if the device is unreachable
            retry_limit = self.retry_limit - 1
        else:
            retry_limit = self.retry_limit

        self.data['info_params'] = self.get_params(params=self.info_params, retry_limit=retry_limit)
        self.data['user_params'] = self.get_params(params=self.user_params)
        self.data['real_values'] = self.get_real_values()
        self.data['memory_status'] = self.memory_status

        return self.data

    def get_params(self, params: list = None, timeout: int = None, retry_limit: float = None):
        for param in params:
            if param not in self.parameters:
                value = None
                try:
                    value = self.read_variable(name=param, timeout=timeout, retry_limit=retry_limit)['data']
                except (InvalidCommand, InvalidAnswer):
                    # Save None to param if an invalid command is executed
                    pass
                except Exception:
                    raise
                self.__update_param(name=param, value=value)
        return {k: v for (k, v) in self.parameters.items() if k in params}

    def get_real_values(self) -> list:
        """ Read real values from device
        Possible output from real values variable:
        - ISM: '7005968.000','0.00','-83.94'
        - RF: '29/08/2020 18:39:47','25.740','3.42'
        """

        rv_mask = [bool(int(i)) for i in self.get_params([RV_MASK])[RV_MASK]]
        rv_units = self.get_params([RV_UNITS])[RV_UNITS].split(';')
        rv_names = self.get_params([RV_NAMES])[RV_NAMES].split(';')
        rv_units = list(compress(rv_units, rv_mask))
        rv_names = list(compress(rv_names, rv_mask))

        # Read real values, add 0.5s to timeout (RV command takes longer to execute)
        real_values = self.read_variable(name=REAL_VALUES, timeout=self.timeout + 0.5,
                                         retry_limit=self.retry_limit)['data']
        if type(real_values) == str:
            real_values = [real_values]
        self.real_values = []

        # First real value can be timestamp, something like 01/01/2020 hh:mm:ss
        # Don't parse timestamp to get python datetime object because timestamp
        # can be even `00/00/2000`. The parser will raise exception
        if len(real_values[0].split('/')) > 2:
            self.real_values.append({'channel': 'Timestamp', 'value': real_values[0], 'unit': None})
            real_values.pop(0)

        for i, val in enumerate(real_values):
            try:
                meas = {
                    'channel': rv_names[i] if len(rv_names) > i else None,
                    'value': val,
                    'unit': rv_units[i] if len(rv_units) > i else None
                }
                self.real_values.append(meas)
            except IndexError as e:
                logger.exception(e)
        return self.real_values

    def get_time(self):
        """Get time from device"""

        return self.read_variable(name=TIME)

    def set_time(self, timestamp: datetime = datetime.now()):
        """Set time of device

        :param timestamp: timestamp to set; if None datetime.now() is used
        :type timestamp: datetime
        """

        if timestamp is None:
            timestamp = datetime.now()
        if type(timestamp) == str:
            try:
                timestamp = datetime.strptime(timestamp.replace(microsecond=0), '%Y-%m-%d %H:%M:%S')
            except ValueError as e:
                logger.error(e)
                raise ValueError('Wrong time format')
        timestamp = str(timestamp.replace(microsecond=0))
        return self.write_variable(name=TIME, value=timestamp)

    def get_memory_status(self) -> dict:
        """Read memory status"""

        if self.rf:
            try:
                data = self.send_receive(app_cmd=MEMORY_STATUS)['data']
                size, free, used = int(data[0]), int(data[1]), int(data[2])
                self.memory_status = {'size': size, 'free': free, 'used': used,
                                      'used_percentage': round(used / size * 100, 1)}
            except NoAnswer:
                raise
            except InvalidAnswer:
                self.memory_status = {'error': True}
            except Exception as e:
                self.memory_status = {'error': True}
                logger.exception(e)
        else:
            self.memory_status = None
        return self.memory_status

    def read_params(self, force=True):
        """Read parameters from device"""

        if self.settings and not force:
            return self.settings
        data = []
        error_count = 0
        max_error_count = 5
        for i in range(500):
            var_name = f'PAR[{i}]'
            try:
                params = self.send_receive(app_cmd=VALUE_READ, var_name=var_name)['data']
                if params == 'End':
                    break
                elif len(params) > 0:
                    params_dict = {}
                    for par in params:
                        if '=' in par:
                            params_dict[par.split('=')[0].lower()] = par.split('=')[1]
                        else:
                            params_dict[par.lower()] = True
                    value = self.read_variable(name=params_dict['name'])['data']
                    params_dict['value'] = value

                    # Extract choices
                    # {'mode': 'T_LIST', 'i0': '0 Enable', 'i1': '1 Disable'}
                    if 'mode' in params_dict:
                        mode = params_dict['mode']
                        if mode == 'T_LIST':
                            choices = []
                            for x in range(0, 50):
                                if f'i{x}' in params_dict:
                                    value = params_dict[f'i{x}'].split(' ')[0]
                                    title = params_dict[f'i{x}'].split(' ')[1]
                                    choices.append({'value': value, 'title': title})
                                else:
                                    break
                            params_dict['choices'] = choices
                    data.append(params_dict)
            except Exception as e:
                logger.warning(f'Cant read param {var_name}, e:{e}')
                error_count += 1
                if error_count == max_error_count:
                    break
                else:
                    continue
        logger.info(f'Settings parameters read, count:{len(data)}')
        self.settings = data
        return data

    def read_memory_data(self) -> bytes:
        """Read data from memory of device and save to memory_data

        :return: Bytes of read data
        """

        self.total_data_read = 0
        self.read_memory_progress = 0
        self.total_data_len = int(self.send_receive(app_cmd=OPEN_MEMORY_DIR)['data'])
        self.raw_memory_data = []
        data_out = b''
        while True:
            data = self.send_receive(app_cmd=MEMORY_START_READING, parse_to_str=False,
                                     stream_channel=True, timeout=2, retry_limit=7)
            # self.raw_memory_data.append(data['data'])
            data['data'] = data_decompress(data['data'])
            self.__update_read_memory_progress(data['data'])
            data_out += data['data']
            if data['status'] == 'OT':
                self.send_receive(app_cmd=MEMORY_END_READING)
                break

        if data_out:
            header, body = data_out.split(b'$HEADER END\r\n')
            header = header.replace(b'\r\n', b'\n') + b'$HEADER END\n'
            self.memory_data = header + body
            return self.memory_data

    def save_memory_data(self, directory='', dtv=True) -> str:
        """Read data from memory of device and save to files.

        :param directory: directory for saving files
        :param dtv: save to dtv format

        Return filepath of created dJ file
        """

        if not self.memory_data:
            raise Exception('Zero data to save')

        ts = datetime.strftime(datetime.now(), "%d%m%Y_%H%M%S")
        filename = f"{self.get_params(['IDENT_A'])['IDENT_A']}_{ts}.jda"
        # Replace all chars that are not allow for path (<>;?\/|*) with empty string
        replacements = ['<', '>', ':', '?', '\\', '*', '|', '/']
        for i in replacements:
            filename = filename.replace(i, '')
        filepath = os.path.join(directory, filename)
        with open(filepath, 'wb') as f:
            f.write(self.memory_data)

        # Convert data to jda format
        # write raw data to file if the conversion was not successful
        try:
            conv = Jda2dtv()
            conv.load_file(filepath)
            if dtv:
                conv.write_dtv_files()
            project_name = self.data['info_params'].get(ID_1, "")
            project_name = project_name.split(': ')[1] if len(project_name.split(': ')) > 1 else project_name
            conv.write_dJ_file(project_name=project_name)
        except Exception as e:
            logger.exception(e)
            filepath = f"raw_{filepath.replace('.jda', '')}"
            with open(filepath, 'a+') as f:
                for data in self.raw_memory_data:
                    f.write(repr(data) + "\n")
            raise
        return filepath.replace('.jda', '.dJ')

    def format_disk(self):
        """Format disk"""

        return self.send_receive(app_cmd=FORMAT_DISK, timeout=self.timeout + 3)

    def export_params(self, filepath: str = 'settings.txt'):
        """Save params to filepath"""

        data = self.read_params(force=False)
        data_write = []
        for i, param in enumerate(data):
            # print (param)
            comment = param['com']
            value = param['value']
            name = param['name']
            data_write.append(f'// Par {i}: {comment}\n')
            data_write.append(f'{name}={value}\n')
        with open(filepath, 'w', encoding='iso-8859-2') as f:
            f.writelines(data_write)
        logger.info(f'Parameters exported: {len(data)}')

    def import_params(self):
        """Import parameters from self.settings to device."""

        if not self.settings:
            raise ParameterError('Zero parameters to import')

        data = self.settings
        failed = []
        for item in data:
            name = item['name']
            value = item['value']
            try:
                self.write_variable(name=name, value=value)
            except Exception as e:
                logger.error(f'Failed to write param {name}, e:{e}')
                failed.append((name, value))

        logger.info(
            f"Parameters imported:{len(data) - len(failed)}{', error:{}'.format(len(failed)) if failed else ''}")
        if failed:
            raise ParameterError(f'Parameters failed to import {[i[0] for i in failed]}')

    def load_params(self, filepath: str) -> list:
        """Load parameters from file into settings variable. Params are loaded not
        actually imported to device. Call import_params for importing this params into device

        :param filepath: Filepath of file with parameters.
        """

        if filepath is None or not os.path.exists(filepath):
            raise FileNotFoundError('File does not exists')

        with open(filepath, encoding='iso-8859-2') as f:
            lines = f.readlines()

        data = []
        for i, line in enumerate(lines):
            if '//' in line:
                try:
                    par = line.split('Par ')[1].split(':')[0]
                    com = "".join(line.split(': ')[1]).strip()
                    line = lines[i + 1]
                    if not re.search('//', line) and re.search('=', line):
                        name = line.split('=')[0].strip()
                        value = line.split('=')[1].strip()
                        data.append({'param': par, 'name': name, 'value': value, 'com': com})
                except IndexError as e:
                    logger.exception(e)
                    pass

        failed = []
        if self.settings:
            for item in data:
                # Replace values with the ones from file
                item_settings = [i for i in self.settings if i['name'] == item['name']]
                if item_settings:
                    item_settings[0]['value'] = item['value']
                else:
                    failed.append(item['name'])
                    logger.error(f'Failed to load param {item["name"]}')
        else:
            self.settings = data
        if failed:
            raise ParameterError(f'Parameters failed to import {[i for i in failed]}')
        return self.settings

    def reset_rf_device(self):
        # !!! Not done
        pass

    def reset_cpu(self):
        """Reset CPU of device"""

        return self.send_receive(app_cmd=RESET_CPU, retry_limit=3)

    def goto_sleep(self):
        """Send RF device to sleep"""

        return self.send_receive(app_cmd=DEVICE_SLEEP)

    def start_wake_up_routine(self):
        """Start wake up procedure for RF devices"""

        if self.rf:
            start = time()
            app_cmd = f"{WAKE_UP_DEVICE}{self.recv_address},{self.wake_up_time}"
            try:
                self.send_receive(app_cmd=app_cmd, retry_limit=3, timeout=0.1)
                sleep(self.wake_up_time)
            except NoAnswer:
                raise NoAnswer('USB router is not working properly')
            wake_time = round(time() - start, 1)
            logger.info(f'Device woke up in {wake_time} seconds')

    def ping_router(self):
        app_cmd = f"{WAKE_UP_DEVICE}1,0"
        return self.send_receive(app_cmd=app_cmd, retry_limit=1, timeout=0.1)

    def connect(self) -> dict:
        """Connect to device (Online mode). If remote device is RF start wake_up procedure."""
        if not self.recv_address:
            raise DeviceNotInitialized('Device is not initialized')

        self.start_wake_up_routine()
        self.get_memory_status()
        data = self.get_data()
        return data

    def disconnect(self, deinit: bool = True):
        """Disconnect from device"""
        if self.rf:
            self.reset_cpu()
        if deinit:
            self.deinitialize()

    def __update_read_memory_progress(self, read_data: bytes):
        """ Update progress in percentage for reading data from device memory

        :param: Read data (one package)
        """
        self.total_data_read += len(read_data)
        self.read_memory_progress = round(self.total_data_read / self.total_data_len * 100, 1)



import sys
from datetime import datetime
from jda_to_dtv import Jda2dtv
from comm_package import replace_repeating_chars, data_decompress


def save_data(data: bytes, filepath='test.jda', dtv=True) -> str:
    with open(filepath, 'wb') as f:
        f.write(data)
    print (f'Jda created: {filepath}')
    conv = Jda2dtv()
    conv.load_file(filepath)
    if dtv:
        conv.write_dtv_files(file_name=filepath.replace('.jda', ''))
    conv.write_dJ_file(file_name=filepath.replace('.jda', '.dJ'))
    return filepath.replace('.jda', '.dJ')


filepath = sys.argv[1]
with open(filepath, 'rb') as f:
    lines = f.readlines()

data = [eval(i) for i in lines]

# Get only data layer from packages. Strip address, confirmation code (e.g. @OT), crc, start-end sign
data_layer = b''
for package in data:
    package = replace_repeating_chars(package, reverse=True)
    package = package[package.find(b'@') + 3:-3]
    package = data_decompress(package)
    data_layer += package

# Split to header and body(actual data) and decompress header
header, body = data_layer.split(b'$HEADER END\r\n')
header = header.replace(b'\r\n', b'\n') + b'$HEADER END\n'

data = header + body
ts = datetime.strftime(datetime.now(), "%d%m%Y_%H%M")
filepath = f'{filepath}_{ts}.jda'
save_data(data, filepath=filepath)



from ctypes import c_ushort

class CRC16(object):
    crc16_tab = []

    # The CRC's are computed using polynomials. Here is the most used
    # coefficient for CRC16
    crc16_constant = 0xA001  # 40961

    def __init__(self, modbus_flag=False):
        # initialize the precalculated tables
        if not len(self.crc16_tab):
            self.init_crc16()
        self.mdflag = bool(modbus_flag)

    def calculate(self, input_data=None):
        try:
            is_string = isinstance(input_data, str)
            is_bytes = isinstance(input_data, bytes)

            if not is_string and not is_bytes:
                raise Exception("Please provide a string or a byte sequence "
                                "as argument for calculation.")

            crcValue = 0x0000 if not self.mdflag else 0xffff

            for c in input_data:
                d = ord(c) if is_string else c
                tmp = crcValue ^ d
                rotated = c_ushort(crcValue >> 8).value
                crcValue = rotated ^ int(self.crc16_tab[(tmp & 0x00ff)], 0)

            return crcValue
        except Exception as e:
            print("EXCEPTION(calculate): {}".format(e))

    def init_crc16(self):
        '''The algorithm uses tables with precalculated values'''
        for i in range(0, 256):
            crc = c_ushort(i).value
            for j in range(0, 8):
                if (crc & 0x0001):
                    crc = c_ushort(crc >> 1).value ^ self.crc16_constant
                else:
                    crc = c_ushort(crc >> 1).value
            self.crc16_tab.append(hex(crc))

def calculate_crc16(data:bytes) -> str:
    """Return calculated crc16 code from string input."""

    crc = CRC16().calculate(data)
    crc_hex = '{:4x}'.format(crc).replace(' ', '0').upper()
    return crc_hex

if __name__ == '__main__':
    input_crc = input('Enter crc input [bytes]:')
    input_crc_bytes = bytes(input_crc, encoding='utf-8')
    crc = calculate_crc16(data=input_crc_bytes)
    print (crc)


from datetime import datetime
import logging
from time import sleep
from serial import SerialException
import re
import os
import string

from .serialib import SerialComm
from .ziva import Ziva

logger = logging.getLogger(__name__)


class SmMonitor(Ziva):
    sm_date = '%d/%m/%Y %H:%M:%S'
    sm_chunk = r'([^=,:\x00]*)'
    sm_pattern = re.compile(r'(?:\s|\x00)*((?:\d|[a-fA-F]){4}):' + sm_chunk + '=' + sm_chunk + ',' + sm_chunk)
    timestamp_format = '%Y-%m-%d %H:%M:%S'

    def __init__(self):
        super().__init__()
        self.running = False
        self.directory = None
        self.data = []
        self.stop_process = False

    def write_dtv(self, device: str, channel: str, value: float, timestamp: datetime, unit: str) -> str:
        """Write data to txt file
        :return: Filepath
        """
        # Create dir d/m/y if it does not exists
        dir_path = os.path.join(self.directory, datetime.strftime(datetime.now(), '%d%m%Y'))
        if not os.path.exists(dir_path):
            os.makedirs(dir_path)

        # create output filepath
        # Remove all hex characters and invalid chars from filename
        filename = f'{device}_{datetime.strftime(datetime.now(), "%d%m%Y")}_{channel.strip()}.txt'
        invalid_chars = '<>:"/\|?*'
        valid_characters = string.printable
        filename = ''.join(i for i in filename if i in valid_characters)
        for char in invalid_chars:
            filename = filename.replace(char, '')
        filepath = os.path.join(dir_path, filename.strip())

        write_data = []
        if not os.path.exists(filepath):
            write_data.append(f'\n\n{unit}\n\n\n\n\n\n\n')

        write_data.append(f'{datetime.strftime(timestamp, "%d/%m/%Y %H:%M:%S")},{value}\n')
        with open(filepath, 'a+') as f:
            f.writelines(write_data)
        logger.debug(f'Data saved to {filepath}: {write_data}')

        return filepath

    def test_router(self):
        """Raise Exception if the router is set to invalid mode (as transmit/receive)"""

        try:
            # If this passes without exceptions it means that the router is in invalid mode
            self.recv_address = 1
            self.ping_router()
            raise ValueError
        except ValueError:
            raise Exception('USB router is not configured properly')
        except Exception as e:
            pass

    def start(self):
        if self.running:
            raise Exception('Already running')

        if not self.directory:
            raise Exception('Directory is not set')

        logger.info('Sm monitor started')

        self.stop_process = False
        self.running = True

        while True:
            if self.stop_process:
                self.running = False
                break

            line = None
            try:
                line = self.readline().strip()
            except (SerialException, AttributeError) as e:
                logger.error(e)
                while True:
                    if self.stop_process:
                        self.running = False
                        break
                    try:
                        self.open()
                        break
                    except SerialException:
                        sleep(2)
            if not line:
                continue
            logger.debug(f"Received: {line}")
            hits = self.sm_pattern.fullmatch(line.decode('latin1'))
            if hits:
                try:
                    device = str(int(hits.group(1), 16))
                    channel = hits.group(2)
                    value = float(hits.group(3))
                    unit = hits.group(4)
                    timestamp = datetime.now().replace(microsecond=0)
                    data = {'device': device, 'channel': channel, 'value': value, 'timestamp': str(timestamp),
                            'unit': unit}
                    filepath = self.write_dtv(device=device, channel=channel,
                                              value=value, timestamp=timestamp, unit=unit)
                    data['filepath'] = filepath
                    for i, item in enumerate(self.data):
                        if item['device'] == device and item['channel'] == channel:
                            self.data[i] = data
                            break
                    else:
                        self.data.append(data)

                except Exception as e:
                    logger.exception(e)
            else:
                logger.warning(f'Invalid data: {line}')

    def stop(self):
        if not self.running:
            raise Exception('Not running')
        self.stop_process = True
        if self.ser is not None:
            self.close_port()
        logger.info('Sm monitor stopped')



import os
import sys

import dateutil.parser
from parser import ParserError
from datetime import datetime


def get_plotly_data(filepaths, parse_filename=True):
    """Get data for plotly from filepaths
    :param parse_filename: Parse filename to device and channel. Use full filename if false
    """

    traces = []
    for file in filepaths:
        if file.endswith('dJ'):
            traces += dj_to_plotly(file)
        if file.endswith('txt') or file.endswith('dtv'):
            traces += dtv_to_plotly(file, parse_filename=parse_filename)
    return traces


def dtv_to_plotly(filepath, parse_filename=True) -> list:
    """ Get plotly ready data from the file of type dtv (can also be txt)"""

    if not filepath:
        raise AttributeError('File is invalid')

    with open(filepath, 'r') as f:
        lines = f.readlines()

    if not filepath.endswith('dtv') and not filepath.endswith('txt'):
        raise AttributeError('File is invalid')

    # Split filepath by `_`. First arg is device, last is channel
    dev_id_split = os.path.splitext(os.path.basename(filepath))[0].split('_')
    if len(dev_id_split) >= 2 or not parse_filename:
        dev_id = f"{dev_id_split[0]}_{dev_id_split[-1]}"
    else:
        dev_id = "".join(dev_id_split)

    data = []
    for line in lines:
        if len(line.split(',')) == 2:
            ts = line.split(',')[0].strip()
            val = line.split(',')[1].strip()
            try:
                dt = datetime.strptime(ts, "%d/%m/%Y %H:%M:%S")
            except ValueError:
                # Invalid datetime string
                continue
            data.append((str(dt), val))

    # Order data by timestamp in ascending order
    data = sorted(data, key=lambda t: t[0])

    # Prepare traces for plotly
    traces = []
    x, y = zip(*data)
    traces = [{
        'x': x,
        'y': y,
        'name': dev_id,
        'type': 'scatter',
        'mode': 'lines+points',
    }]

    return traces


def dj_to_plotly(filepath) -> list:
    """Get plotly ready data from the file of type `dJ`"""

    if not filepath:
        raise AttributeError('File is invalid')

    with open(filepath, 'r') as f:
        lines = f.readlines()

    if not filepath.endswith('dJ'):
        raise AttributeError('File is invalid')

    header = {}
    for line in lines:
        if line.find('=') > -1:
            header[line.split('=')[0]] = line.split('=')[1].strip()
        elif line.strip() == '$DATA':
            break

    data = {}
    for line in lines:
        if len(line.split(',')) >= 2:
            ts = line.split(',')[0]
            values = line.split(',')[1:]
            try:
                dt = dateutil.parser.parse(ts)
            except ParserError:
                # Invalid datetime string
                continue
            for i, val in enumerate(values):
                try:
                    val = val.strip()
                    val = eval(val)
                except (NameError, SyntaxError):
                    # Invalid val
                    continue
                if f'CHAN_NAME{i}' not in data:
                    data[f'CHAN_NAME{i}'] = []
                data[f'CHAN_NAME{i}'].append((str(dt), val))

    # Order data by timestamp in ascending order
    for key, val in data.items():
        data[key] = sorted(val, key=lambda t: t[0])

    # Prepare data for plotly
    traces = []
    for key, val in data.items():
        x, y = zip(*val)
        traces.append({
            'x': x,
            'y': y,
            'name': header[key],
            'type': 'scatter',
            'mode': 'lines+points',
        })
    return traces


if __name__ == '__main__':
    file = sys.argv[1]
    print(get_plotly_data([file]))



"""Communication exceptions"""


class NotConnectedToSerial(Exception):
    """If connection to serial port was not established"""


class NoAnswer(Exception):
    """There is no answer from a device."""


class InvalidAnswer(Exception):
    """Answer was received but it is now a valid one."""


class CRCError(Exception):
    """CRC Error"""


class InvalidCommand(Exception):
    """Command/Message that was sent is not a valid one"""


class DeviceNotInitialized(Exception):
    """Device is not configured"""


class ParameterError(Exception):
    """One or multiple parameters failed to read/write"""


class DeviceTimeout(Exception):
    """Device timeout"""


class InvalidData(Exception):
    """Invalid data"""

class InvalidAddress(Exception):
    """Invalid address of the device"""



from datetime import datetime
from random import random
from time import sleep
import os

from ziva.const import RV_MASK, WAKE_UP_TIME_TOTAL, REAL_VALUES
from ziva.exceptions import InvalidCommand
from ziva.ziva import Ziva


class ZivaMock(Ziva):
    def __init_(self):
        super().__init__()

    def initialize(self, recv_address: int, rf: bool = True, wake_up_time: int = WAKE_UP_TIME_TOTAL,
                   user_params: list = None):
        mock_file = 'mock_rf.txt' if rf else 'mock_ism.txt'
        self.read_mock_settings(os.path.join(os.path.dirname(os.path.realpath(__file__)), mock_file))
        super().initialize(recv_address=recv_address, rf=rf, wake_up_time=wake_up_time, user_params=user_params)

    def read_mock_settings(self, filepath):
        with open(filepath, 'r', encoding='iso-8859-2') as f:
            lines = f.readlines()
        self.settings = []

        com = None
        for line in lines:
            if '//' in line and len(line.split(':')) == 2:
                com = line.split(':')[1].strip()
            if '//' not in line and len(line.split('=')) == 2:
                name = line.split('=')[0].strip()
                value = line.split('=')[1].strip()
                self.parameters[name] = value
                param = {
                    'name': name,
                    'title': name,
                    'com': com,
                    'value': value,
                    'mode': 'T_STRING'
                }
                self.settings.append(param)

    def write_variable(self, name=None, value=None, timeout: float = None, retry_limit: int = None):
        name, value = str(name), str(value)
        if name in self.parameters:
            self.parameters[name] = value
        else:
            raise NotImplemented

    def read_variable(self, name: str, timeout: float = None, retry_limit: int = None):
        data_out = {'data': None, 'status': 'ok'}
        if name == REAL_VALUES:
            data = []
            if self.rf:
                data.append(datetime.strftime(datetime.now(), "%d/%m/%Y %H:%M:%S"))
            for i in self.parameters[RV_MASK]:
                if bool(int(i)):
                    data.append(str(round(random()*100, 1)))
            data_out['data'] = data
        else:
            if name not in self.parameters:
                raise InvalidCommand
            data_out['data'] = self.parameters[name]
        return data_out

    def get_memory_status(self) -> dict:
        if self.rf:

            size, free, used = 1000, 500, 500
            self.memory_status = {'size': size, 'free': free, 'used': used,
                                  'used_percentage': round(used / size * 100, 1)}
        else:
            self.memory_status = None
        return self.memory_status

    def start_wake_up_routine(self):
        if self.rf:
            sleep(self.wake_up_time)

    def reset_cpu(self):
        return True

    def read_params(self, force=True):
        return super().read_params(force=False)

    def set_time(self, timestamp: datetime = datetime.now()):
        return True



""" Communication constants"""


# Structure commands
START_MARKER = '82'
END_MARKER = '83'
START_MARKER_BYTES = b'\x82'
END_MARKER_BYTES = b'\x83'
SERVER_ADDR_BYTES = b'\x00\x00\x00\x00'
EQUALS_SIGN = '{:X}'.format(ord('='))

# Single application commands
VALUE_READ = 'VR'  # Value read
VALUE_WRITE = 'VW'  # Value write
REAL_VALUES = 'RV'  # Value read


# NAME OF PARAMS
IDENT_A = 'IDENT_A'
IDENT_B = 'IDENT_B'
IDENT_C = 'IDENT_C'
ID_0 = 'ID[0]'
ID_1 = 'ID[1]'
RV_MASK = 'RV_MASK'
RV_UNITS = 'RV_UNITS'
RV_NAMES = 'RV_NAMES'
RECORD_TIME = 'RECORD_TIME'
TIME = 'TIME'

WAKE_UP_DEVICE = 'SCWD' # wake up device
DEVICE_SLEEP = 'SCN'
RESET_CPU = 'SRESET'

# Memory operations
OPEN_MEMORY_DIR = 'COFRData0.jda'
MEMORY_START_READING = 'CTFR'
MEMORY_END_READING = 'CCFR'
MEMORY_STATUS = 'FDSX'
FORMAT_DISK = 'FFX0'

# retry
RETRY_NO_ANSWER = 3
RETRY_CRC_ERROR = 10
RETRY_INVALID_ANSWER = 10
RETRY_INVALID_COMMAND = 0

WAKE_UP_TIME_STEP = 30
WAKE_UP_TIME_TOTAL = 30




# ==============================================================================
#    OBJEKT ZA DELO Z MICRA DATOTEKAMI ".jda"
#
#    Avtor:   JAZON
#    Datum:   8-12-2020
#    Verzija: 1.3
# ==============================================================================

import datetime
import sys


class Jda2dtv:
    # Objekt za pretvorbo "jda" datoteke v "dJ" in "dtv" datoteko
    def __init__(self):
        self.chanels = []
        self.ch_sizes = []
        self.records = []

    def load_file(self, file_name):
        # Glavna funcija za pretvorbo. Vraca: true/false OK/neuspeh
        inx = file_name.find('.')
        self.raw_file_name = file_name[:inx]
        with open(file_name, "rb") as file:
            self.buff = file.read()
        lt = self.buff.decode('iso-8859-2')
        self.data_index = lt.find('$HEADER END')    # + cr, lf 0
        self.header = lt[:self.data_index+11]
        while self.buff[self.data_index] != 0:
            self.data_index += 1
        self.data_index += 1
        self.convert_header()
        type = self.dheader['FILE_TYPE']
        if type == '0' or type == '1':
            raise Exception("Napaka 3: Datoteki tipa 0 in 1 nista podprti")
        self.file_type_23()

    def convert_header(self):
        # Pretvorba headerja v dictionary
        self.dheader = {}
        for line in self.header.splitlines():
            inx = line.find('=')
            if inx > 0:
                key = line[:inx]
                val = line[inx+1:]
                self.dheader[key] = val

    def file_type_23(self):
        # Kreira listo rekordov iz podatkov datoteke tipa 2 in 3
        self.get_chan_data()
        self.records = []
        for i in range(self.no_records):
            self.records.append(self.get_record())

    def buff_num(self, n_bytes):
        # Pretvorba n_bytes iz buff v cifro + premakne pointer
        x = 0
        for i in range(n_bytes):
            bi = n_bytes - i - 1
            x += (self.buff[self.data_index] << (8*bi))
            self.data_index += 1
        return x

    def buff_chan(self):
        ''' Pretvori no_data_bytes: izpod pointerja v listo bitnih vrednosti kanalov
            Primer:
            ch_sizes = [15, 6] -> dolzina podatkov v rekordu je 3 byti

            bits:     23 22 21 20 19 18 17 16 15 14 13 12 11 10 9 8 7 6 5 4 3 2 1 0
            ch0bits:                                                    5 4 3 2 1 0
            ch1bits:        15 14 13 12 11 10  9  8  7  6  5  4 3 2 1 0

            + premakne pointer
            return: listo bitnih vrednosti kanalov
        '''
        no_shifts = (self.no_data_bytes * 8) - sum(self.ch_sizes)
        bits = 0
        for i in range(self.no_data_bytes):
            bits = bits << 8
            bits += self.buff[self.data_index + i]
        bits = bits >> no_shifts
        chans = []
        for no_bits in reversed(self.ch_sizes):
            mask = (2 ** no_bits) - 1
            chans.append(bits & mask)
            bits = bits >> no_bits
        self.data_index += self.no_data_bytes
        chans.reverse()
        return chans

    def get_no_records(self):
        # Izmeri stevilo rekordov
        no_data_bytes = sum(self.ch_sizes) / 8            # Stevilo bytov podatkov kanalov
        if(sum(self.ch_sizes) % 8) != 0:
            no_data_bytes += 1
        self.no_data_bytes = int(no_data_bytes)
        record_size = self.no_data_bytes + 4                   # Stevilo bytov za cas
        data_size = len(self.buff) - self.data_index
        if (data_size % record_size) != 0:
            raise Exception("Napaka 2: Nepravilna velikost podatkov")
        self.no_records = int(data_size / record_size)

    def get_raw_record(self):
        # Iz buffra vzame en record - vrne surove cifre izpod pointerja premakne pointer v bufru
        time = self.buff_num(4)
        ch_data = self.buff_chan()
        return time, ch_data

    def get_record(self):
        # Vrne formatiran preracunan rekord v listo: [cas, kanal, kanal, ... ]
        time, data = self.get_raw_record()
        record = [self.num_time(time)]
        for i,chan in enumerate(self.chanels):
            tmin = float(chan['MIN'])
            tmax = float(chan['MAX'])
            max_val = (2 ** self.ch_sizes[i]) - 1
            val = data[i] * (tmax - tmin) / max_val + tmin
            format = self.chanels[i]['FOR']
            sval = self.chanels[i]['FOR'] % val
            record.append(sval)
        return record

    def num_time(s, num):
        # Pretvorba MICRA time v PYTHON datetime
        dni = int(num / 86400)
        leto = int(dni / 372) + 2000
        dni %= 372
        mesec = int(dni / 31) + 1
        date = dni % 31 + 1
        num  %= 86400
        hours = int(num / 3600)
        num %= 3600
        minutes = int(num / 60)
        seconds = num % 60
        time = datetime.datetime(leto, mesec, date, hours, minutes, seconds)
        return time

    def get_chan_data(self):
        # Ekstrahira podatke kanalov: Prisotnost, NAME, UNIT, FOR, MIN, MAX
        self.chan_mask = 0
        self.chanels = []
        lb = self.dheader['CH_BITS']
        if lb[-1] == ',':
            lb = lb[:-1]
        self.ch_sizes = lb.split(",")
        self.ch_sizes = [int(i) for i in self.ch_sizes]
        for ch in range(8):
            key = "CH" + str(ch) + "_NAME"
            if key in self.dheader:
                self.chan_mask |= 0x01 << ch
                chan = {}
                chan['INDEX'] = ch
                chan['NAME'] = self.dheader[key]
                key = "CH" + str(ch) + "_UNIT"
                chan['UNIT'] = self.dheader[key]
                key = "CH" + str(ch) + "_FOR"
                chan['FOR'] = self.dheader[key]
                key = "CH" + str(ch) + "_MIN"
                chan['MIN'] = self.dheader[key]
                key = "CH" + str(ch) + "_MAX"
                chan['MAX'] = self.dheader[key]
                self.chanels.append(chan)
        self.get_no_records()

    def write_dJ_file(self, file_name='', project_name=""):
        # Kreiranje dJ datoteke
        if file_name == '':
            file_name = self.raw_file_name + '.dJ'
        self.tex = ''
        self.add_line('NO_CHAN=%d' % len(self.chanels))
        for i, chan in enumerate(self.chanels):
            self.add_line('CHAN_NAME%d=%s'   % (i, chan['NAME']))
            self.add_line('CHAN_UNIT%d=%s'   % (i, chan['UNIT']))
            self.add_line('CHAN_FORMAT%d=%s' % (i, chan['FOR']))
            self.add_line('CHAN_MIN%d=%s'    % (i, chan['MIN']))
            self.add_line('CHAN_MAX%d=%s'    % (i, chan['MAX']))
        self.add_line('PROJECT=' + project_name)
        self.add_line('$DATA')
        for record in self.records:
            for i,ch in enumerate(record):
                if i == 0:
                    lt = record[0].strftime("%Y/%m/%d %H:%M:%S")
                else:
                    if i > 1:
                        lt += ',' + record[i]
                    else:
                        lt += ', ' + record[i]
            self.add_line(lt + ',')
        with open(file_name, "w", encoding='iso-8859-2') as text_file:
            text_file.write(self.tex)

    def write_dtv_files(self, file_name=''):
        # Kreiranje Henrik (txt) datoteke
        # Datoteke dobijo ime  file_name_(ImeKanala).txt
        # Če parametra ni je file_name ime izvorne datoteke
        if file_name == '':
            file_name = self.raw_file_name
        for i, chan in enumerate(self.chanels):
            self.tex = '\n\n' + chan['UNIT'] + '\n' * 8
            for record in self.records:
                self.add_line(record[0].strftime("%d/%m/%Y %H:%M:%S") + ',' + record[i+1])
            fname = file_name + '_' + chan['NAME'] + '.txt'
            with open(fname, "w", encoding='iso-8859-2') as text_file:
                text_file.write(self.tex)

    def add_line(self, text):
        self.tex += text + '\n'


if __name__ == '__main__':
    filename = sys.argv[1]
    conv = Jda2dtv()
    conv.load_file(file_name=filename)
    conv.write_dJ_file()
    conv.write_dtv_files()
